////////////////////////////////////////////////
//          REAL TIME HDR RENDERING
// A program testing HDR tonemapping on modern graphic 
// hardware.
// Only yet tested on Nvidia card
//
//              Jonas Nilsson
//              Camilla Kempainen
////////////////////////////////////////////////

//C++ package to be used
#include <GL/glew.h>
#include <GL/glut.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "textfile.h"//Functions to read shader textfile
#include "readRAW.h"//Functions to read RAW format
#include "HDRloader.h"//Functions to read hdr format
#include "rgbe.h"

#define MEAN 1
#define SCURVE 2
#define BLOOM 3
#define RETINEX 4

//Start window size
static int winXsize = 640;
static int winYsize = 480;
int frametexX = winXsize;
int frametexY = winYsize;

//Fps counter variables
float t0 = 0.0;
float FPSmin=10000;
float FPSmax=0;
float fps=0;
int frames = 0;
int totframes =0;
char titlestring[200];
GLfloat time = 0.0;

//Variables for interaction
int mousex,mousey;
bool leftmouse,middlemouse;
static float zoom = 4;
static float rot = 90;
static float rot2 = 90;
static float transy = 0;
static float transx = 0;
static GLfloat size = 1;
static int object = 1;
static int loop = 0;
static int timeloop = 1;
static int lightrot= 0;
static int texture = 1;
static int shaderopt = 0;
static float exposure = 1.0;
static int tonemapping = MEAN;

//Menu variables
static int textmenu;
static int mainmenu;
static int shadermenu;

//Gluint object used in the program
GLuint envTextureID;
GLuint tonemapping_program;
GLuint RGBEenc_program;
GLuint filter_program;
GLuint thesphere;
GLuint frameTextureID;
GLuint fbobject;

GLuint blurtexture1;
GLuint blurtexture2;
GLuint blurtexture3;
GLuint blurtexture4;
GLuint blurtexture5;
GLuint blurtexture6;
float f = 2.0;

float w[6];

//Error handling variables
GLenum errcode;
char errstring[20000];

//Funktions used in the program
void setShaderProgram(GLuint * program, char * shadername);
void key(unsigned char key, int x, int y);
void menu(int value);
void createMenu(void);
static void resize(int width, int height);
void showFPS();
static void display(void);
static void idle(void);
void drawInversedTexturedSphere(float r, int segs);
void createRAWtext(GLuint *texid,char* name,int width,int height);
void createHDRtext(GLuint *texid,const char* name);
void shadermenufunction(int value);
void texmenufunc(int value);
void menu(int value);

void calcW()
{
    float sum = 0.0f;
    for(int i=0;i<6;i++)
    {
      sum += pow(6-i+1,f);            
    }
    for(int i=0;i<6;i++)
    {
      w[i] = pow(6-i+1,f) / sum;
    }
}

////////////////////////////////////////////////
//               SHADERS
//
////////////////////////////////////////////////
//Loader shader program
//Report errrors
void setShaderProgram(GLuint * program, char * vertname,char * fragname)
{
char *vs,*fs;
*program = glCreateProgram();
GLint shaderstatus;
//Loading and compiling vertedx shader	
		GLuint vertex_shader = glCreateShader(GL_VERTEX_SHADER);
	    vs = textFileRead(vertname);
		const char * vv = vs;
  		glShaderSource(vertex_shader, 1, &vv,NULL);
	    free(vs);
       	glCompileShader(vertex_shader);
		glAttachShader(*program,vertex_shader);
       glGetShaderiv(vertex_shader,GL_COMPILE_STATUS,&shaderstatus);
       if(shaderstatus != GL_TRUE)
       {
         glGetShaderInfoLog(vertex_shader,sizeof(errstring),NULL,errstring);
         MessageBox (NULL, (char*)errstring,"Vertex shader error", 0);
       }    
//Loading and compiling fragment shader
        GLuint fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);	
		fs = textFileRead(fragname);
		const char * ff = fs;
		glShaderSource(fragment_shader, 1, &ff,NULL);
        free(fs);
		glCompileShader(fragment_shader);
    	glAttachShader(*program,fragment_shader);
       glGetShaderiv(fragment_shader,GL_COMPILE_STATUS,&shaderstatus);
       if(shaderstatus != GL_TRUE)
       {
         glGetShaderInfoLog(fragment_shader,sizeof(errstring),NULL,errstring);
         MessageBox (NULL, (char*)errstring,"Fragment shader error", 0);
       }
		glLinkProgram(*program);
        glUseProgram(*program);
        
}






////////////////////////////////////////////////
//               DISPLAY
//
////////////////////////////////////////////////

//Resize function
//The setup for camera and viewport
static void 
resize(int width, int height)
{
    const float ar = (float) width / (float) height;
    
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
        // Set viewport. This is the pixel rectangle we want to draw into.
    glViewport( 0, 0, width, height ); // The entire window

    // Select and setup the projection matrix.
    glMatrixMode(GL_PROJECTION); // "We want to edit the projection matrix"
    glLoadIdentity(); // Reset the matrix to identity
    // 45 degrees FOV, same aspect ratio as viewport, depth range 1 to 100
    gluPerspective( 70.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.1f );

    // Select and setup the modelview matrix.
    glMatrixMode( GL_MODELVIEW ); // "We want to edit the modelview matrix"
    glLoadIdentity(); // Reset the matrix to identity
    // Look from 0,-4,0 towards 0,0,0 with Z as "up" in the image
    
    gluLookAt( 0.0f, -4.0f, 0.0f,  // Eye position
               0.0f, 0.0f, 0.0f,   // View point
               0.0f, 0.0f, 1.0f ); // Up vector
}


//Fps and other information is written out in the titlebar.
void showFPS() {
  // If one second has passed, or if this is the very first frame
    if( (time-t0) > 1.0 || totframes ==0)
    {
        fps = (float)frames / (time-t0);
        if(totframes>10)
        if(fps<FPSmin) FPSmin=fps;
        if(fps>FPSmax) FPSmax=fps;
        
        t0 = time;

        frames = 0;

    }
      sprintf(titlestring, "Shadertest %.1f FPS (Max %.1f, Min %.1f), %.1f seconds and %.1d frames elapsed. Retinex F %.1f", fps,FPSmax,FPSmin,time,totframes,f);
        glutSetWindowTitle(titlestring); 
  
    if(loop) rot += 0.2;
    frames ++;
    totframes+=frames;
}

/*
 * drawTexturedSphere(r, segs) - Draw a sphere centered on the local
 * origin, with radius "r" and approximated by "segs" polygon segments,
 * with texture coordinates in a latitude-longitude mapping.
 */
 //Orginal made by Stefan Gustavsson. I've just inverted the normals to create a 
 //environment sphere.
void drawInversedTexturedSphere(float r, int segs) {
  int i, j;
  float x, y, z, z1, z2, R, R1, R2;

  // Top cap
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0,0,1);
  glTexCoord2f(0.5f,1.0f); // This is an ugly (u,v)-mapping singularity
  glVertex3f(0,0,r);
  z = cos(M_PI/segs);
  R = sin(M_PI/segs);
    for(i = 0; i <= 2*segs; i++) {
      x = R*cos(i*2.0*M_PI/(2*segs));
      y = R*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(-x, -y, -z);
      glMultiTexCoord2f(GL_TEXTURE0,(float)i/(2*segs), 1.0f-1.0f/segs);
      glVertex3f(r*x, r*y, r*z);
    }
  glEnd();  

  // Height segments
  for(j = 1; j < segs-1; j++) {
    z1 = cos(j*M_PI/segs);
    R1 = sin(j*M_PI/segs);
    z2 = cos((j+1)*M_PI/segs);
    R2 = sin((j+1)*M_PI/segs);
    glBegin(GL_TRIANGLE_STRIP);
    for(i = 0; i <= 2*segs; i++) {
      x = R1*cos(i*2.0*M_PI/(2*segs));
      y = R1*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(-x, -y, -z1);
      glMultiTexCoord2f(GL_TEXTURE0,(float)i/(2*segs), 1.0f-(float)j/segs);
      glVertex3f(r*x, r*y, r*z1);
      x = R2*cos(i*2.0*M_PI/(2*segs));
      y = R2*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(-x, -y, -z2);
      glTexCoord2f((float)i/(2*segs), 1.0f-(float)(j+1)/segs);
      glVertex3f(r*x, r*y, r*z2);
    }
    glEnd();
  }

  // Bottom cap
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0,0,-1);
  glTexCoord2f(0.5f, 1.0f); // This is an ugly (u,v)-mapping singularity
  glVertex3f(0,0,-r);
  z = -cos(M_PI/segs);
  R = sin(M_PI/segs);
    for(i = 2*segs; i >= 0; i--) {
      x = R*cos(i*2.0*M_PI/(2*segs));
      y = R*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(-x, -y, -z);
      glMultiTexCoord2f(GL_TEXTURE0,1.0f-(float)i/(2*segs), 1.0f/segs);
      glVertex3f(r*x, r*y, r*z);
    }
  glEnd();
}
/*
 * initSphereList(GLuint *listID, GLdouble scale) - create a display list
 * to render the sphere more efficently than calling lots of trigonometric
 * functions for each frame.
 * (A vertex array would be even faster, but I'm a bit lazy here.)
 */
void initSpheresList(GLuint *listID, GLdouble scale)
{
  *listID = glGenLists(1);
  
  glNewList(*listID, GL_COMPILE);
  drawInversedTexturedSphere(scale,20);

  glEndList();
}

//Drawing a textured rectangle.
void drawTexturedRectangle(float x1,float y1,float x2,float y2)
{
  glBegin(GL_POLYGON); 
  glMultiTexCoord2f(GL_TEXTURE0,0.0,0.0); glVertex2f(x1, y1);
  glMultiTexCoord2f(GL_TEXTURE0,1.0,0.0); glVertex2f(x2, y1);
  glMultiTexCoord2f(GL_TEXTURE0,1.0,1.0); glVertex2f(x2, y2);
  glMultiTexCoord2f(GL_TEXTURE0,0.0,1.0); glVertex2f(x1, y2);
  glEnd();
}
void blurTexture(GLuint * text,int nr,int w,int h)
{
     glActiveTexture(GL_TEXTURE0+2);
     glBindTexture(GL_TEXTURE_2D,frameTextureID);
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     glUniform1i(glGetUniformLocation(filter_program,"w"),w);
     glUniform1i(glGetUniformLocation(filter_program,"h"),h);
     drawTexturedRectangle(0,0,w,h);

     glActiveTexture(GL_TEXTURE0+nr);
     glBindTexture(GL_TEXTURE_2D,*text);
     glGenerateMipmapEXT(GL_TEXTURE_2D);
     glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, w, h,0);
}
//The display function, run in each frame.
static void 
display(void)
{   // glEnable(GL_CULL_FACE);
glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glUniform1f(glGetUniformLocation(tonemapping_program,"sizeY"),glutGet(GLUT_WINDOW_HEIGHT));
    glUniform1f(glGetUniformLocation(tonemapping_program,"sizeX"),glutGet(GLUT_WINDOW_WIDTH));
     
    if(texture) glEnable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE0+1);
    glBindTexture(GL_TEXTURE_2D, envTextureID);
   
    time = glutGet(GLUT_ELAPSED_TIME) / 1000.0;    
   // glEnable(GL_LIGHTING);
    glLoadIdentity();
    glPushMatrix();

    glColor3d(0.5,0.5,0.5);
    glTranslatef(transy*0.1,transx*0.1,-zoom*0.1);     

    glRotatef(rot, 0.0f, 1.0f, 0.0f );
    glRotatef(rot2, 1.0f, 0.0f, 0.0f );   

   if(shaderopt) glUseProgram(RGBEenc_program); 
     glUniform1i(glGetUniformLocation(RGBEenc_program,"envTexture2"),1);
 
    glCallList(thesphere);
   
    //Render the polygon with framebuffer as texture
    if(object)
    {
     glActiveTexture(GL_TEXTURE0+2);
     glBindTexture(GL_TEXTURE_2D,frameTextureID);
     glGenerateMipmapEXT(GL_TEXTURE_2D);
     glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0,  glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT),0);

     glUseProgram(filter_program);
     glUniform1i(glGetUniformLocation(filter_program,"texture"),2);

     glLoadIdentity();
     glMatrixMode(GL_PROJECTION);
     glPushMatrix();
     glLoadIdentity();
     gluOrtho2D(0.0,glutGet(GLUT_WINDOW_WIDTH),0.0,glutGet(GLUT_WINDOW_HEIGHT));

     if(tonemapping==RETINEX)
     {     
     blurTexture(&blurtexture1,3,16,16);
     blurTexture(&blurtexture2,4,32,32);
     blurTexture(&blurtexture3,5,64,64);
     blurTexture(&blurtexture4,6,128,128);
     blurTexture(&blurtexture5,7,256,256);
     blurTexture(&blurtexture6,8,512,512);
     
     glUseProgram(tonemapping_program);    
    
 
     glUniform1f(glGetUniformLocation(tonemapping_program,"w1"),w[0]);
     glUniform1f(glGetUniformLocation(tonemapping_program,"w2"),w[1]);
     glUniform1f(glGetUniformLocation(tonemapping_program,"w3"),w[2]);
     glUniform1f(glGetUniformLocation(tonemapping_program,"w4"),w[3]);
     glUniform1f(glGetUniformLocation(tonemapping_program,"w5"),w[4]);
     glUniform1f(glGetUniformLocation(tonemapping_program,"w6"),w[5]);
     
     glUniform1i(glGetUniformLocation(tonemapping_program,"blur1"),3);
     glUniform1i(glGetUniformLocation(tonemapping_program,"blur2"),4);
     glUniform1i(glGetUniformLocation(tonemapping_program,"blur3"),5);
     glUniform1i(glGetUniformLocation(tonemapping_program,"blur4"),6);
     glUniform1i(glGetUniformLocation(tonemapping_program,"blur5"),7);
     glUniform1i(glGetUniformLocation(tonemapping_program,"blur6"),8);
     }
     else if(tonemapping==BLOOM)
     {
     blurTexture(&blurtexture1,3,32,32);
     glUseProgram(tonemapping_program);    
     glUniform1i(glGetUniformLocation(tonemapping_program,"smalltexture"),3);    
     }
     else glUseProgram(tonemapping_program);
     
     glActiveTexture(GL_TEXTURE0+2);
     glBindTexture(GL_TEXTURE_2D,frameTextureID);
     glUniform1i(glGetUniformLocation(tonemapping_program,"envTexture"),2);
     
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     drawTexturedRectangle(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);

   
    }
    glUniform1f(glGetUniformLocation(tonemapping_program,"exposure"),exposure); 
    errcode=glGetError();
    glutSwapBuffers();
    
    glDisable(GL_TEXTURE_2D);
}

//Idle function
static void 
idle(void)
{
 showFPS();
 glutPostRedisplay();
}




////////////////////////////////////////////////
//                  INERACTIONS
//
////////////////////////////////////////////////


//Menu option function
//Each menu item corresponds to an integer.
void menu(int value)
{
  if(value == 0)
  {
    
    exit(0);
  }
  else if(value==1)
  {
                 if (loop==1)
             loop=0;
             else loop=1;
  }
  else if(value==2)
  {
                 if (object==1)
             object=0;
             else object=1;
  }
  else if(value==3)
  {
  if(shaderopt)
    {
    shadermenufunction(tonemapping);
    }
  }
  
}
//Submenu option function.
//Each menu item corresponds to an integer.
void texmenufunc(int value)
{
 switch(value)
 {
              case 1 :
                   //createRAWtext(&envTextureID,"RAW/doge2.raw",4096,2048);    
                   createHDRtext(&envTextureID,"HDRimages/doge2.hdr");
              case 2 :
                   //createRAWtext(&envTextureID,"RAW/ennis.raw",3072,1536);
                   createHDRtext(&envTextureID,"HDRimages/ennis.hdr");
                   break;
              case 3 :
                   //createRAWtext(&envTextureID,"RAW/glacier.raw",2048,1024);
                   createHDRtext(&envTextureID,"HDRimages/glacier.hdr");
                   break;
              case 4 :
                   //createRAWtext(&envTextureID,"RAW/grace-new.raw",3072,1536);
                   createHDRtext(&envTextureID,"HDRimages/grace-new.hdr");
                   break;
              case 5 :
                   //createRAWtext(&envTextureID,"RAW/pisa.raw",4096,2048);
                   createHDRtext(&envTextureID,"HDRimages/pisa.hdr");
                   break;
              case 6 :
                   //createRAWtext(&envTextureID,"RAW/uffizi-large.raw",2400,1200);
                   createHDRtext(&envTextureID,"HDRimages/uffizi-large.hdr");
                   break;
 }
              
}
//Shadermenu options
void shadermenufunction(int value)
{
     switch(value){
                  case 1 :
                       setShaderProgram(&tonemapping_program,"shader.vert","mean.frag");
                       tonemapping = MEAN;
                       break;
                  case 2 : 
                       setShaderProgram(&tonemapping_program,"shader.vert","scurve.frag");
                       tonemapping = SCURVE;
                       break;
                  case 3 :
                       setShaderProgram(&tonemapping_program,"shader.vert","bloom.frag");
                       tonemapping = BLOOM;
                       break;
                  case 4 :
                       setShaderProgram(&tonemapping_program,"shader.vert","retinex.frag");
                       tonemapping = RETINEX;
                       break;
                       }    
     
}

//Creation of the menu.
void createMenu(void){
  textmenu = glutCreateMenu(texmenufunc);
  //glutAddMenuEntry("doge2, 100MB", 1);
  glutAddMenuEntry("ennis", 2);
  glutAddMenuEntry("glacier", 3);
  glutAddMenuEntry("grace-new", 4);
  glutAddMenuEntry("pisa", 5);
  glutAddMenuEntry("uffizi-large", 6);
  
  shadermenu = glutCreateMenu(shadermenufunction);
  glutAddMenuEntry("Normalized by mean",1);
  glutAddMenuEntry("S-curve",2);
  glutAddMenuEntry("Bloom",3);
  glutAddMenuEntry("Rahman Retinex",4);

  // Create the menu, this menu becomes the current menu
  mainmenu = glutCreateMenu(menu);
  
  // Create an entry
  glutAddMenuEntry("Auto rotation", 1);

  glutAddMenuEntry("Wndow", 2);
  
  glutAddSubMenu("Texture", textmenu);
  if(shaderopt) glutAddSubMenu("Tonemapping", shadermenu);
  // Create an entry
  if(shaderopt) glutAddMenuEntry("Reload shader",3);
  glutAddMenuEntry("Quit", 0);

  // Let the menu respond on the right mouse button
  glutAttachMenu(GLUT_RIGHT_BUTTON);
}
//Mouse interaction function.
//Runs when a mouse button is clicked
void
mouse(int button, int state, int x, int y)
{

     mousex=x;
     mousey=y;
     
     leftmouse=(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN);

     middlemouse=(button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN);     
     
}
//Mouse motion function
//Runs when the mouse is moved over the window.
void processMouseActiveMotion(int x, int y)
{
     if(leftmouse)
     {
     rot+=((x-mousex)*0.5);                
     rot2+=((mousey-y)*0.5);    
     }
     else if(middlemouse)
     {
     zoom-=((mousey-y)*0.1);    
     }
     mousex=x;
     mousey=y;
     glutPostRedisplay();
}
//Keyhandler. Handles input from the keyboard.
void key(unsigned char key, int x, int y)
{
    switch (key) 
    {
        case 27 : 
             exit(0);
             break;
        case 'w':
             if (object==1)
             object=0;
             else object=1;
             break;
       case 'e':
             if (loop==1)
             loop=0;
             else loop=1;
             break;
      case 'r':
             if (timeloop==1)
             timeloop=0;
             else timeloop=1;
             break;
      case 't':
             if (lightrot==1)
             lightrot=0;
             else lightrot=1;
             break;
             
        case '-':
            if(tonemapping==RETINEX)
                  {
                  f-=0.1;
                  calcW();
                  }
            else exposure+=0.1;
            break;
        case '+':
            if(tonemapping==RETINEX)
                   {
                   f+=0.1;
                   calcW();
                   }
            else exposure-=0.1;
            break;
        
            
        case '9':
            transx++;
            break;
        case '3':
            transx--;
            break;
        case '*':
            transy++;
            break;
        case '/':
            transy--;
            break;

        case '7':
            size++;
            break;
        case '1':
            size--;
            break;
    }

    glutPostRedisplay();
}




////////////////////////////////////////////////
//               TEXTURES
//
////////////////////////////////////////////////

//Creating an empty texture
void createTexture(GLuint *texture,int xSize, int ySize)
{
 //new array
//float* colorBits = new float[ xSize * ySize * 3 ];

//texture creation..
glGenTextures(1,texture);
glBindTexture(GL_TEXTURE_2D,*texture);

glTexImage2D(GL_TEXTURE_2D,0 ,4 , xSize,
	ySize, 0 , GL_RGBA, GL_FLOAT, NULL);

//you can set other texture parameters if you want
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);

//glGenerateMipmapEXT(GL_TEXTURE_2D);
//clean up
//delete[] colorBits;
}

//Creating an float texture from RAW file
void createRAWtext(GLuint *texid,char* name,int width,int height)
{

    glBindTexture(GL_TEXTURE_2D,*texid);
    
    try
    {    
    float* data = readRAWfile(name,width,height);
    
    
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB16F_ARB, width, height, 0, GL_RGB, GL_FLOAT, data);
    delete data;
    }
    catch(int)
    {
      MessageBox (NULL, "Couldn't read texture" , "Error", 0);  
      }       
}

//Create a RGBE texture from an hdr file
void createHDRtext(GLuint *texid,const char* name)
{
     glBindTexture(GL_TEXTURE_2D,*texid);
        
     FILE * file  = fopen(name,"rb");;
     		
     int w,h;
     
     if(RGBE_ReadHeader(file, &w, &h, NULL)==RGBE_RETURN_FAILURE)
         MessageBox (NULL, "Couldn't read file" , "Error", 0);  
     
     float * data = (float*)malloc(sizeof(float) * w * h *3); 
     
     if(RGBE_ReadPixels_RLE(file, data, w, h)==RGBE_RETURN_FAILURE)
     MessageBox (NULL, "Couldn't read texture" , "Error", 0);  
      
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB16F_ARB, w, h, 0, GL_RGB, GL_FLOAT, data);
    delete data;
 
     /*
    HDRLoaderResult HDRdata;
    
    glBindTexture(GL_TEXTURE_2D,*texid);
        
    if(!HDRload(name, HDRdata)) MessageBox (NULL, "Couldn't read texture" , "Error", 0);
    
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    //glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    //glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);    
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, HDRdata.width, HDRdata.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, HDRdata.cols);
    delete &HDRdata;
    */
}







/* Program entry point */
int 
main(int argc, char *argv[])
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_ALPHA);
    glutInitWindowSize(winXsize,winYsize);
    glutInitWindowPosition(10,10);
    resize(winXsize,winYsize);

    glutCreateWindow("");

    glutReshapeFunc(resize);
    glutDisplayFunc(display);
    glutKeyboardFunc(key);
    glutIdleFunc(idle);
    glutMouseFunc(mouse);
    glutMotionFunc(processMouseActiveMotion);
    glClearColor(0,0,0,1);
   
    glCullFace(GL_BACK);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    
    glEnable(GL_LIGHT0);
    glEnable(GL_NORMALIZE);
    glEnable(GL_COLOR_MATERIAL);
    glDisable(GL_BLEND);

    glewInit();
   	if (glewIsSupported("GL_VERSION_2_0"))
   	{
		shaderopt = 1;
    }	
	else 
    {
	    shaderopt = 0;
		MessageBox (NULL, "No Opengl 2.0 support, shaders can not be used" , "Error", 0);
	}
    	
	  initSpheresList(&thesphere, 10.0);
//Generating hdr environment texture
    glGenTextures(1,&envTextureID);
    //createRAWtext(&envTextureID,"RAW/ennis.raw",3072,1536);
    createHDRtext(&envTextureID,"HDRimages/ennis.hdr");

//generate frame texture    
    createTexture(&frameTextureID, frametexX,frametexY);
    
    createTexture(&blurtexture1, 64,64);
    createTexture(&blurtexture2, 64,64);
    createTexture(&blurtexture3, 64,64);
    createTexture(&blurtexture4, 64,64);
    createTexture(&blurtexture5, 64,64);
    createTexture(&blurtexture6, 64,64);
    
    calcW();
    
//Loading shaders, if supported
    if(shaderopt) {
                   setShaderProgram(&tonemapping_program,"shader.vert","mean.frag");
                   setShaderProgram(&RGBEenc_program,"shader.vert","RGBEenc.frag");
                   setShaderProgram(&filter_program,"shader.vert","filter.frag");
                   glUniform1f(glGetUniformLocation(tonemapping_program,"exposure"),exposure);
                   }
 
    createMenu();  
    
    glutMainLoop();

    return EXIT_SUCCESS;
}

